'''
Provides a reader and writer for TSV files, similar to the CSV implementation.
'''

from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import TextIO

__version__ = '0.0.1'


def escape(cell: Optional[str]) -> str:
    '''Make a string unambiguous when used in a tsv'''
    if cell is None:
        return r"\N"

    return (cell
            .replace("\\", r"\B")  # nonterminal symbol
            .replace("\n", r"\n")
            .replace("\r", r"\r")
            .replace("\t", r"\t")
            .replace(r"\B", "\\\\")
            )


def unescape(cell: str) -> Optional[str]:
    '''Deserialize a string from the tsv storage format'''

    if cell == r"\N":
        return None

    return (cell
            .replace(r"\\", r"\B")  # nonterminal symbol
            .replace(r"\n", "\n")
            .replace(r"\r", "\r")
            .replace(r"\t", "\t")
            .replace(r"\B", "\\")
            )


class Reader:
    '''Reads and unexcapes rows of tsv line by line.'''

    def __init__(self, f: Iterable[str]):
        self._file_it: Iterator[str] = iter(f)
        self._width = None

    def __iter__(self):
        return self

    def __next__(self) -> List[str]:
        '''Gets the next non-empty row of values.'''
        row: str = ""
        while not row:
            row = next(self._file_it).strip("\n").strip("\r")  # use removesuffix in 3.9

        assert "\n" not in row
        assert "\r" not in row
        values = row.split("\t")

        self._expect_number_of_cells_to_be_constant(len(values))

        return list(map(unescape, values))

    def _expect_number_of_cells_to_be_constant(self, width):
        '''Check that the file isn't ragged.'''
        if self._width is not None:
            if width is not self._width:
                raise ValueError("Cannot read rows with different widths.")
        else:
            self._width = width


class Writer:
    '''Writes rows into a tsv file. Performs escaping.'''

    def __init__(self, f: TextIO):
        self._file: TextIO = f
        self._width = None

    def write_row(self, cells: Iterable[Optional[str]]) -> None:
        '''Writes a single row into a tsv file. Performs escaping.'''
        self._expect_number_of_cells_to_be_constant(len(cells))

        escaped_cells = (escape(cell) for cell in cells)
        row = "\t".join(escaped_cells) + "\n"
        self._file.write(row)

    def _expect_number_of_cells_to_be_constant(self, width):
        '''Check that the file isn't ragged.'''
        if self._width is not None:
            if width is not self._width:
                raise ValueError("Cannot write rows with different widths.")
        else:
            self._width = width


class DictReader:
    '''Allows reading of tsv files in the form of dicts.'''

    def __init__(self,
                 f: Iterable[str],
                 fieldnames: Optional[List[str]] = None):
        """Allows reading of tsv files in the form of dicts.

        Args:
            f (Iterable[str]):
                The file to be read.
            fieldnames (List[str], optional):
                The keys to be assigned each row (and the resulting dict).
                If omitted, the keys are read from the first row.
        """
        self._reader: Reader = Reader(f)
        self.fieldnames: List[str] = fieldnames or next(self._reader)

    def __iter__(self):
        return self

    def __next__(self) -> Dict[str, str]:
        cells = next(self._reader)
        if len(cells) != len(self.fieldnames):
            raise ValueError(
                "The number of cells in the current line "
                "doesn't match the number of fieldnames. \n",
                {
                    "fieldnames": self.fieldnames,
                    "row": cells
                })
        return dict(zip(self.fieldnames, cells))


class DictWriter:
    '''Allows writing of dict values to a new or existing tsv file.'''

    def __init__(self,
                 f: TextIO,
                 write_header: bool,
                 fieldnames: Optional[Sequence[str]] = None,
                 missing_values_placeholder: Optional[str] = None,
                 allow_excess_values: bool = False):
        """Allows writing of dict values to a new or existing tsv file.

        Args:
            f (TextIO):
                The file to be written to.
            write_header (bool, optional):
                Whether the fieldnames should be written back to the file.
            fieldnames (Sequence[str], optional):
                The columns and their order.
                By default, they are inferred from the first dict passed.
            missing_values_placeholder (str, optional):
                Used when a dict is missing a value for a column.
                By default, throw a ValueException.
            allow_excess_values (bool, optional):
                Should dict entries without a column be discarded silently?
                By default, throw a KeyException.
        """
        self._expect_bool_type(write_header, "write_header")
        self._expect_bool_type(allow_excess_values, "allow_excess_values")
        self._writer = Writer(f)
        self._write_header = write_header
        self.fieldnames: Sequence[str] = fieldnames or None
        self.missing_values_placeholder: str = missing_values_placeholder
        self.allow_excess_values: bool = allow_excess_values

        if write_header and fieldnames is not None:
            self._writer.write_row(fieldnames)
            self._write_header = False

    @staticmethod
    def _expect_bool_type(value, name: str):
        if value not in (True, False):
            raise TypeError(f"{name} is expected to be a bool.")

    def write_row(self, rowdict: Dict[str, Any]) -> None:
        '''Writes a dict into the TSV as a single line'''

        if self.fieldnames is None:
            self.fieldnames = rowdict.keys()

        if self._write_header:
            self._writer.write_row(self.fieldnames)
            self._write_header = False

        row = self._build_row(rowdict)

        self._writer.write_row(row)

    def _build_row(self, rowdict: Dict[str, Any]) -> List[Optional[str]]:
        '''Constructs a row to match the file width'''

        if not self.allow_excess_values:
            self._expect_no_excess_values(rowdict)

        if self.missing_values_placeholder is None:
            self._expect_no_missing_values(rowdict)

        row: List[Optional[str]] = []
        for key in self.fieldnames:
            if key in rowdict:
                row.append(rowdict[key])
            else:
                row.append(self.missing_values_placeholder)
        return row

    def _expect_no_excess_values(self, rowdict):
        '''Throws an exception if values are not written back'''
        excess_values = set(rowdict.keys()) - set(self.fieldnames)
        if excess_values:
            raise KeyError(
                "These rowdict keys could not be written to the file: "
                + str(excess_values) + "\n"
                "Consider setting allow_excess_values in init."
            )

    def _expect_no_missing_values(self, rowdict):
        '''Throws an exception if values missing'''
        missing_values = set(self.fieldnames) - set(rowdict.keys())
        if missing_values:
            raise ValueError(
                "Rowdict is missing a value for rows: "
                + str(missing_values) + "\n"
                "Consider setting a missing_values_placeholder in init.")

    def write_rows(self, rowdicts: Iterable[Dict[str, Any]]):
        '''Adds several dicts in the form of as many rows'''
        for rowdict in rowdicts:
            self.write_row(rowdict)



from tsv2dict import DictReader
from tsv2dict import DictWriter
from tsv2dict import Reader
from tsv2dict import Writer
from tsv2dict import escape
from tsv2dict import unescape


class MockFile():
    def __init__(self):
        self.content: str = ""

    def write(self, s: str):
        self.content += s

    def readlines(self):
        return self.content.split('\n')


unescaped1 = "Harry\nAnd the chamber of pythons\r:"
escaped1 = r"Harry\nAnd the chamber of pythons\r:"
unescaped2 = "'\t' is escaped as '\\t'."
escaped2 = r"'\t' is escaped as '\\t'."


def test_escape():
    assert escape(unescaped1) == escaped1
    assert escape(unescaped2) == escaped2


def test_unescape():
    assert unescape(escaped1) == unescaped1
    assert unescape(escaped2) == unescaped2


test_tsv = [
    "Name\tAge\tAddress\n",
    "Me\t\tHere\n",
    "Jens\t5\t1234\\tVessy\n",
    "Jim\t23\tKreis 4\\nZürich\n",
    "\n",
    "The sun\t4'603'000'000\tn\\\\a\n",
]

test_header = ["Name", "Age", "Address"]
test_dicts = [
    {"Name": "Me", "Age": "", "Address": "Here"},
    {"Name": "Jens", "Age": "5", "Address": "1234\tVessy"},
    {"Name": "Jim", "Age": "23", "Address": "Kreis 4\nZürich"},
    {"Name": "The sun", "Age": "4'603'000'000", "Address": "n\\a"},
]
test_tsv_file = "tests/contacts.tsv"
test_tsv_lines = [test_header] + [list(d.values()) for d in test_dicts]


def test_reader():
    reader = Reader(test_tsv)
    actual = next(reader)
    assert actual == test_header


def test_reader_with_file():
    with open(test_tsv_file) as f:
        reader = Reader(f)
        for expected in test_tsv_lines:
            actual = next(reader)
            assert actual == expected


def test_reader_expects_equal_number_of_cells_per_line():
    f = ["a\tb\n",
         "1\t2\n",
         "1\t2\t3\n"]
    reader = Reader(f)
    next(reader)
    next(reader)
    try:
        next(reader)
        assert False
    except ValueError:
        assert True


def test_writer():
    f = MockFile()
    writer = Writer(f)
    writer.write_row(test_header)
    assert f.content == test_tsv[0]


def test_writer_expects_equal_number_of_cells_per_line():
    f = MockFile()
    writer = Writer(f)
    writer.write_row(["a", "b"])
    writer.write_row(["1", "2"])
    try:
        writer.write_row(["1", "2", "3"])
        assert False
    except ValueError:
        assert True


def test_dict_reader():
    reader = DictReader(test_tsv)
    assert reader.fieldnames == ["Name", "Age", "Address"]
    first = next(reader)
    assert first["Name"] == "Me"
    second = next(reader)
    assert second["Age"] == "5"
    third = next(reader)
    assert third["Address"] == "Kreis 4\nZürich"


def test_dict_reader_implements_iterable():
    dicts = [dict_ for dict_ in DictReader(test_tsv)]
    assert len(dicts) is len(test_dicts)


def test_dict_writer_with_write_header():
    f = MockFile()
    _ = DictWriter(f, True, ["A", "B", "C"])
    assert "A\tB\tC\n" == f.content


def test_values_are_written_in_the_same_order_as_the_header():
    f = MockFile()
    writer = DictWriter(f, False, ["B", "A", "C"])
    writer.write_row({"C": "c", "B": "2", "A": "1.00"})
    assert f.content == "2\t1.00\tc\n"


def test_writing_escapes_reserved_characters():
    f = MockFile()
    writer = DictWriter(f, False, ["B", "A", "C"])
    unescaped = "Harry\nAnd the chamber of pythons\\:"
    escaped = r"Harry\nAnd the chamber of pythons\\:"
    writer.write_row({"C": unescaped, "B": "2", "A": "1.00"})
    assert f"2\t1.00\t{escaped}\n" == f.content



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup

# with open("README.md") as f:
#     long_description = f.read()


def read(*names, **kwargs):
    with io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fh:
        return fh.read()


setup(
    name='tsv2dict',
    version='0.0.1',
    license='MIT',
    description='Python support for linear TSV files',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    author='Nico Kurmann',
    author_email='nico.kurmann@gmail.com',
    url='https://github.com/nkurmann/tsv2dict',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        "Operating System :: OS Independent",
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        # uncomment if you test on these interpreters:
        # 'Programming Language :: Python :: Implementation :: IronPython',
        # 'Programming Language :: Python :: Implementation :: Jython',
        # 'Programming Language :: Python :: Implementation :: Stackless',
        'Topic :: Utilities',
    ],
    project_urls={
        'Documentation': 'https://tsv2dict.readthedocs.io/',
        'Changelog': 'https://tsv2dict.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/nkurmann/tsv2dict/issues',
    },
    keywords=[
        'tsv',
        'persistence',
        'export'
    ],
    python_requires='>=3.6',
    install_requires=[
        # eg: 'aspectlib==1.1.1', 'six>=1.7',
    ],
    extras_require={
        # eg:
        #   'rst': ['docutils>=0.11'],
        #   ':python_version=="2.6"': ['argparse'],
        "dev": [
            "autopep8",
            "binaryornot",
            "check-manifest",
            "click",
            "flake8",
            "jinja2-time",
            "mypy",
            "pip-chill",
            "poyo",
            "pylint",
            "pytest",
            "python-slugify",
            "tox",
            "tsv2dict",
            "twine",
        ]
    },
    entry_points={
        'console_scripts': [
            'tsv2dict = tsv2dict.cli:main',
        ]
    },
)
